import datetime

import pytest

from phantom.predicates.datetime import is_tz_aware
from phantom.predicates.datetime import is_tz_naive

parametrize_aware = pytest.mark.parametrize(
    "dt", (datetime.datetime.now(tz=datetime.timezone.utc),)
)
parametrize_naive = pytest.mark.parametrize(
    "dt", (datetime.datetime.now(), datetime.datetime(1969, 12, 23))
)


class TestIsTZAware:
    @parametrize_aware
    def test_returns_true_for_aware_dt(self, dt: datetime.datetime) -> None:
        assert is_tz_aware(dt) is True

    @parametrize_naive
    def test_returns_false_for_naive_dt(self, dt: datetime.datetime) -> None:
        assert is_tz_aware(dt) is False


class TestIsTZNaive:
    @parametrize_naive
    def test_returns_true_for_naive_dt(self, dt: datetime.datetime) -> None:
        assert is_tz_naive(dt) is True

    @parametrize_aware
    def test_returns_false_for_aware_dt(self, dt: datetime.datetime) -> None:
        assert is_tz_naive(dt) is False



from typing import Any
from typing import Pattern

from .base import Phantom
from .predicates import re


class Match(str, Phantom, abstract=True):
    def __init_subclass__(cls, pattern: Pattern[str], **kwargs: Any) -> None:
        super().__init_subclass__(predicate=re.is_match(pattern), **kwargs)


class FullMatch(str, Phantom, abstract=True):
    def __init_subclass__(cls, pattern: Pattern[str], **kwargs: Any) -> None:
        super().__init_subclass__(predicate=re.is_full_match(pattern), **kwargs)



from phantom import Phantom
from phantom.predicates import boolean


class P:
    ...


class Q:
    ...


class R(P, Q):
    ...


class A(P, Phantom, predicate=boolean.true):
    ...


class B(Q, Phantom, predicate=boolean.true):
    ...


class C(A, B):
    ...


print(C.__predicate__)
print(C.__bound__)



from .base import Predicate
from .boolean import negate
from .generic import equal


def less(n: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return value < n

    return check


def le(n: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return value <= n

    return check


def greater(n: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return value > n

    return check


def ge(n: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return value >= n

    return check


positive = greater(0)
non_positive = le(0)
negative = less(0)
non_negative = ge(0)


def modulo(n: float, p: Predicate[float]) -> Predicate[float]:
    def check(value: float) -> bool:
        return p(value % n)

    return check


even: Predicate[int] = modulo(2, equal(0))
odd: Predicate[int] = negate(even)



import pytest

from phantom.ext.phonenumbers import FormattedPhoneNumber
from phantom.ext.phonenumbers import InvalidPhoneNumber
from phantom.ext.phonenumbers import PhoneNumber
from phantom.ext.phonenumbers import _deconstruct_phone_number
from phantom.ext.phonenumbers import is_formatted_phone_number
from phantom.ext.phonenumbers import is_phone_number
from phantom.ext.phonenumbers import normalize_phone_number


class TestPhoneNumber:
    def test_unparsable_number_is_not_instance(self):
        assert not isinstance("+46", PhoneNumber)

    def test_invalid_number_is_not_instance(self):
        assert not isinstance("+467012345678", PhoneNumber)

    def test_unformatted_number_is_instance(self):
        assert isinstance("+46 (701) 234567", PhoneNumber)


class TestFormattedPhoneNumber:
    def test_unparsable_number_is_not_instance(self):
        assert not isinstance("+46", FormattedPhoneNumber)

    def test_invalid_number_is_not_instance(self):
        assert not isinstance("+467012345678", FormattedPhoneNumber)

    def test_unformatted_number_is_not_instance(self):
        assert not isinstance("+46 (701) 234567", FormattedPhoneNumber)

    def test_formatted_number_is_instance(self):
        assert isinstance("+46701234567", FormattedPhoneNumber)

    def test_normalizes_unformatted_number(self):
        number = FormattedPhoneNumber.parse("+46 (701) 234567")
        assert number == "+46701234567"
        assert isinstance(number, FormattedPhoneNumber)

    def test_parse_raises_for_invalid_phone_number(self):
        with pytest.raises(InvalidPhoneNumber):
            FormattedPhoneNumber.parse("+46")

    def test_raises_type_error_for_out_of_bound_type(self):
        """Since we override parse we need to test the bound check"""
        value = 123
        with pytest.raises(
            TypeError, match=fr"Can't parse str from int value: {value}"
        ):
            FormattedPhoneNumber.parse(123)


class TestDeconstructPhoneNumber:
    def test_can_parse_international_phone_number_without_country_code(self):
        result = _deconstruct_phone_number("123456789", "SE")
        assert result.country_code == 46
        assert result.national_number == 123456789

    def test_can_parse_international_phone_number_with_country_code(self):
        result = _deconstruct_phone_number("+46123456789", "NO")
        assert result.country_code == 46
        assert result.national_number == 123456789

    def test_can_parse_national_phone_number_with_country_code(self):
        result = _deconstruct_phone_number("0123456789", "SE")
        assert result.country_code == 46
        assert result.national_number == 123456789

    def test_raises_invalid_phone_number_for_insufficient_country_data(self):
        with pytest.raises(InvalidPhoneNumber) as exc_info:
            _deconstruct_phone_number("0701")
        assert exc_info.value.error_type == InvalidPhoneNumber.INVALID_COUNTRY_CODE

    def test_raises_invalid_phone_number_for_parse_exception(self):
        with pytest.raises(InvalidPhoneNumber) as exc_info:
            _deconstruct_phone_number("0")
        assert exc_info.value.error_type == InvalidPhoneNumber.NOT_A_NUMBER

    def test_raises_invalid_phone_number_for_invalid_phone_number(self):
        with pytest.raises(InvalidPhoneNumber) as exc_info:
            _deconstruct_phone_number("+461234567890")
        assert exc_info.value.error_type == InvalidPhoneNumber.INVALID


class TestNormalizePhoneNumber:
    def test_can_normalize_national_number_with_country_code(self):
        assert normalize_phone_number("(123) 456 789", "SE") == "+46123456789"

    def test_can_normalize_international_number_without_country_code(self):
        assert normalize_phone_number("+46 (123) 456 789") == "+46123456789"


class TestIsPhoneNumber:
    def test_returns_true_for_valid_number(self):
        assert is_phone_number("+46 (123) 456789") is True

    def test_returns_false_for_invalid_number(self):
        assert is_phone_number("+461234567890") is False


class TestIsFormattedPhoneNumber:
    def test_returns_true_for_formatted_number(self):
        assert is_formatted_phone_number("+46123456789") is True

    def test_returns_false_for_unformatted_number(self):
        assert is_formatted_phone_number("+46 (123) 456 789") is False



import datetime

from .base import Phantom
from .predicates.datetime import is_tz_aware
from .predicates.datetime import is_tz_naive


class TZAware(Phantom, bound=datetime.datetime, predicate=is_tz_aware):
    ...


class TZNaive(Phantom, bound=datetime.datetime, predicate=is_tz_naive):
    ...



# Workaround for something that looks like this bug
# https://github.com/pytest-dev/pytest/issues/4386
from __future__ import annotations

import functools
from typing import Any
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union


class UnresolvedClassAttribute(NotImplementedError):
    ...


def resolve_class_attr(
    cls: type, name: str, argument: Optional[object], required: bool = True
) -> None:
    argument = getattr(cls, name, None) if argument is None else argument
    if argument is not None:
        setattr(cls, name, argument)
    elif required and not getattr(cls, "__abstract__", False):
        raise UnresolvedClassAttribute(
            f"Concrete phantom type {cls.__qualname__} must define class attribute "
            f"{name}."
        )


A = TypeVar("A")


def excepts(
    exception: Union[Tuple[Type[Exception], ...], Type[Exception]],
    negate: bool = False,
) -> Callable[[Callable[[A], Any]], Callable[[A], bool]]:
    """Turn a unary function that raises an exception into a boolean predicate."""

    def decorator(fn: Callable[[A], Any]) -> Callable[[A], bool]:
        @functools.wraps(fn)
        def wrapper(arg: A, /) -> bool:
            try:
                fn(arg)
            except exception:
                return negate
            return not negate

        return wrapper

    return decorator



from .base import Phantom
from .base import PhantomBase
from .base import Predicate
from .base import parse_bound

__version__ = "0.5.0"
__all__ = ("PhantomBase", "Phantom", "Predicate", "parse_bound")



import datetime

from .boolean import negate


def is_tz_aware(dt: datetime.datetime) -> bool:
    # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
    return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None


is_tz_naive = negate(is_tz_aware)



import pytest

from phantom.ext.iso3166 import Alpha2
from phantom.ext.iso3166 import InvalidCountryCode
from phantom.ext.iso3166 import normalize_alpha2_country_code


class TestNormalizeAlpha2CountryCode:
    @pytest.mark.parametrize(
        "given, normalized",
        (("sE", "SE"), ("dk", "DK"), ("IE", "IE")),
    )
    def test_normalizes_mixed_case_valid_country_code(
        self, given: str, normalized: str
    ) -> None:
        assert normalize_alpha2_country_code(given) == normalized

    @pytest.mark.parametrize("invalid", ("UK", "not a country code"))
    def test_raises_for_invalid_country_code(self, invalid: str) -> None:
        with pytest.raises(InvalidCountryCode):
            normalize_alpha2_country_code(invalid)


class TestAlpha2:
    @pytest.mark.parametrize("invalid", ("SP", "DA", "AV", 1))
    def test_invalid_country_code_is_not_instance(self, invalid: object) -> None:
        assert not isinstance(invalid, Alpha2)

    @pytest.mark.parametrize("country_code", ("PS", "AD", "VA"))
    def test_valid_country_code_is_instance(self, country_code: str) -> None:
        assert isinstance(country_code, Alpha2)

    def test_normalizes_valid_country_code(self) -> None:
        country_code = Alpha2.parse("ps")
        assert country_code == "PS"
        assert isinstance(country_code, Alpha2)

    @pytest.mark.parametrize("invalid", ("SP", "DA", "AV"))
    def test_raises_for_invalid_country_code(self, invalid: str) -> None:
        with pytest.raises(InvalidCountryCode):
            Alpha2.parse(invalid)



import datetime

import pytest

from phantom.datetime import TZAware
from phantom.datetime import TZNaive

parametrize_aware = pytest.mark.parametrize(
    "dt", (datetime.datetime.now(tz=datetime.timezone.utc),)
)
parametrize_naive = pytest.mark.parametrize(
    "dt", (datetime.datetime.now(), datetime.datetime(1969, 12, 23))
)


class TestTZAware:
    @parametrize_aware
    def test_aware_datetime_is_instance(self, dt):
        assert isinstance(dt, TZAware)

    @parametrize_naive
    def test_naive_datetime_is_not_instance(self, dt):
        assert not isinstance(dt, TZAware)

    @parametrize_naive
    def test_instantiation_raises_for_naive_datetime(self, dt):
        with pytest.raises(TypeError):
            TZAware.parse(dt)

    @parametrize_aware
    def test_instantiation_returns_instance(self, dt):
        assert dt is TZAware.parse(dt)


class TestTZNaive:
    @parametrize_naive
    def test_naive_datetime_is_instance(self, dt):
        assert isinstance(dt, TZNaive)

    @parametrize_aware
    def test_aware_datetime_is_not_instance(self, dt):
        assert not isinstance(dt, TZNaive)

    @parametrize_aware
    def test_instantiation_raises_for_aware_datetime(self, dt):
        with pytest.raises(TypeError):
            TZNaive.parse(dt)

    @parametrize_naive
    def test_instantiation_returns_instance(self, dt):
        assert dt is TZNaive.parse(dt)



from typing import Tuple

import pytest

from phantom.predicates import interval

Boundaries = Tuple[float, float]
parametrize_open_edge = pytest.mark.parametrize(
    "value, boundaries",
    [
        (1, (1, 2)),
        (2, (1, 2)),
        (1.0, (1.0, 2)),
        (2.0, (1, 2.0)),
    ],
)
parametrize_closed_edge = pytest.mark.parametrize(
    "value, boundaries",
    [
        (1.0001, (1, 2)),
        (1.9999, (1, 2)),
        (1.00001, (1.0, 2)),
        (1.99999, (1, 2.0)),
    ],
)
parametrize_closed_outside = pytest.mark.parametrize(
    "value, boundaries",
    [
        (0, (1, 2)),
        (3, (1, 2)),
        (0.99999, (1.0, 2)),
        (2.00001, (1, 2.0)),
        (20, (1, 3)),
    ],
)


class TestOpen:
    def test_returns_true_for_middle_value(self) -> None:
        assert interval.open(1, 3)(2) is True

    @parametrize_open_edge
    def test_returns_true_for_inclusive_edge_value(
        self, value: float, boundaries: Boundaries
    ) -> None:
        assert interval.open(*boundaries)(value) is True

    @parametrize_closed_outside
    def test_returns_false_for_exclusive_outside_value(
        self, value: float, boundaries: Boundaries
    ) -> None:
        assert interval.open(*boundaries)(value) is False


class TestClosed:
    def test_returns_true_for_middle_value(self) -> None:
        assert interval.closed(1, 3)(2) is True

    @parametrize_open_edge
    def test_returns_false_for_inclusive_edge_value(
        self, value: float, boundaries: Boundaries
    ) -> None:
        assert interval.closed(*boundaries)(value) is False

    @parametrize_closed_outside
    def test_returns_false_for_exclusive_outside_value(
        self, value: float, boundaries: Boundaries
    ) -> None:
        assert interval.closed(*boundaries)(value) is False

    @parametrize_closed_edge
    def test_returns_true_for_exclusive_edge_value(
        self, value: float, boundaries: Boundaries
    ) -> None:
        assert interval.closed(*boundaries)(value) is True


class TestOpenClosed:
    def test_returns_true_for_middle_value(self) -> None:
        assert interval.open_closed(1, 3)(2) is True

    def test_lower_bound(self) -> None:
        intv = interval.open_closed(5, 10)
        assert intv(5) is True
        assert intv(4.9999) is False

    def test_upper_bound(self) -> None:
        intv = interval.open_closed(5, 10)
        assert intv(10) is False
        assert intv(9.9999) is True


class TestClosedOpen:
    def test_returns_true_for_middle_value(self) -> None:
        assert interval.closed_open(1, 3)(2) is True

    def test_lower_bound(self) -> None:
        intv = interval.closed_open(5, 10)
        assert intv(5.0001) is True
        assert intv(5) is False

    def test_upper_bound(self) -> None:
        intv = interval.closed_open(5, 10)
        assert intv(10) is True
        assert intv(10.00001) is False



from typing import Iterable

import pytest

from phantom.base import Predicate
from phantom.predicates import boolean


class TestTrue:
    @pytest.mark.parametrize("value", [0, False, 1, "test", ()])
    def test_returns_true_for_any_given_value(self, value: object) -> None:
        assert boolean.true(value) is True


class TestFalse:
    @pytest.mark.parametrize("value", [0, False, 1, "test", ()])
    def test_returns_false_for_any_given_value(self, value: object) -> None:
        assert boolean.false(value) is False


parametrize_truthy = pytest.mark.parametrize("value", [1, "a", (0,), [""], True])
parametrize_falsy = pytest.mark.parametrize("value", [0, "", (), [], False])


class TestTruthy:
    @parametrize_truthy
    def test_returns_true_for_truthy_value(self, value: object) -> None:
        assert boolean.truthy(value) is True

    @parametrize_falsy
    def test_returns_false_for_falsy_value(self, value: object) -> None:
        assert boolean.truthy(value) is False


class TestFalsy:
    @parametrize_falsy
    def test_returns_true_for_falsy_value(self, value: object) -> None:
        assert boolean.falsy(value) is True

    @parametrize_truthy
    def test_returns_false_for_truthy_value(self, value: object) -> None:
        assert boolean.falsy(value) is False


class TestBoth:
    def test_returns_true_for_two_succeeding_predicates(self) -> None:
        assert boolean.both(boolean.true, boolean.true)(0) is True

    def test_returns_false_for_failure_in_first_predicate(self) -> None:
        assert boolean.both(boolean.false, boolean.true)(0) is False

    def test_returns_false_for_failure_in_second_predicate(self) -> None:
        assert boolean.both(boolean.true, boolean.false)(0) is False


parametrize_all_true = pytest.mark.parametrize(
    "predicates",
    [
        (boolean.true,),
        (boolean.true, boolean.true),
        (boolean.true, boolean.true, boolean.true),
    ],
)
parametrize_some_false = pytest.mark.parametrize(
    "predicates",
    [
        (boolean.true, boolean.false),
        (boolean.false, boolean.true),
        (boolean.true, boolean.false, boolean.true),
    ],
)
parametrize_all_false = pytest.mark.parametrize(
    "predicates",
    [
        (boolean.false,),
        (boolean.false, boolean.false),
        (boolean.false, boolean.false, boolean.false),
    ],
)


class TestAllOf:
    @parametrize_all_true
    def test_returns_true_for_succeeding_predicates(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.all_of(predicates)(0) is True

    @parametrize_some_false
    def test_returns_false_for_some_failing_predicate(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.all_of(predicates)(0) is False

    @parametrize_all_false
    def test_returns_false_for_only_failing_predicate(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.all_of(predicates)(0) is False


class TestAnyOf:
    @parametrize_all_true
    def test_returns_true_for_succeeding_predicates(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.any_of(predicates)(0) is True

    @parametrize_some_false
    def test_returns_true_for_some_failing_predicate(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.any_of(predicates)(0) is True

    @parametrize_all_false
    def test_returns_false_for_only_failing_predicate(
        self, predicates: Iterable[Predicate]
    ) -> None:
        assert boolean.any_of(predicates)(0) is False



from .base import Predicate


def open(low: float, high: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return low <= value <= high

    return check


def open_closed(low: float, high: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return low <= value < high

    return check


def closed_open(low: float, high: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return low < value <= high

    return check


def closed(low: float, high: float) -> Predicate[float]:
    def check(value: float) -> bool:
        return low < value < high

    return check



from typing import Tuple
from typing import Union

from .base import Predicate


def equal(a: object) -> Predicate[object]:
    def check(b: object) -> bool:
        return a == b

    return check


def identical(a: object) -> Predicate[object]:
    def check(b: object) -> bool:
        return a is b

    return check


def of_type(t: Union[type, Tuple[type, ...]]) -> Predicate[object]:
    def check(a: object) -> bool:
        return isinstance(a, t)

    return check



from __future__ import annotations

from typing import Final
from typing import cast

import iso3166

from phantom import Phantom
from phantom import parse_bound
from phantom.predicates.collection import contained

__all__ = (
    "ALPHA2",
    "is_alpha2_country_code",
    "InvalidCountryCode",
    "Alpha2",
    "CountryCode",
    "normalize_alpha2_country_code",
)

ALPHA2: Final = frozenset(iso3166.countries_by_alpha2.keys())
is_alpha2_country_code = contained(ALPHA2)


class InvalidCountryCode(TypeError):
    ...


def normalize_alpha2_country_code(country_code: str) -> Alpha2:
    normalized = country_code.upper()
    if not is_alpha2_country_code(normalized):
        raise InvalidCountryCode
    return cast(Alpha2, normalized)


class Alpha2(str, Phantom, predicate=is_alpha2_country_code):
    @classmethod
    def parse(cls, instance: object) -> Alpha2:
        return normalize_alpha2_country_code(parse_bound(str, instance))


CountryCode = Alpha2



from typing import Container
from typing import Iterable
from typing import Sized

import pytest

from phantom.base import Predicate
from phantom.predicates import collection
from phantom.predicates import generic
from phantom.predicates import numeric


class TestContains:
    @pytest.mark.parametrize(
        "item, container",
        [
            (1, (1, 2)),
            ("b", "abc"),
        ],
    )
    def test_returns_true_for_container_with_item(
        self, item: object, container: Container
    ) -> None:
        assert collection.contains(item)(container) is True

    @pytest.mark.parametrize(
        "item, container",
        [
            (1, (2, 3)),
            ("d", "abc"),
        ],
    )
    def test_returns_false_for_container_without_item(
        self, item: object, container: Container
    ) -> None:
        assert collection.contains(item)(container) is False


class TestContained:
    @pytest.mark.parametrize(
        "container, item",
        [
            ((1, 2), 1),
            ("abc", "b"),
        ],
    )
    def test_returns_true_for_item_in_container(
        self, container: Container, item: object
    ) -> None:
        assert collection.contained(container)(item) is True

    @pytest.mark.parametrize(
        "container, item",
        [
            ((2, 3), 1),
            ("abc", "d"),
        ],
    )
    def test_returns_false_for_item_not_in_container(
        self, container: Container, item: object
    ) -> None:
        assert collection.contained(container)(item) is False


class TestCount:
    @pytest.mark.parametrize(
        "predicate, sized",
        [
            (generic.equal(1), [0]),
            (numeric.le(3), ("a", "b")),
            (numeric.ge(3), "abc"),
        ],
    )
    def test_returns_true_for_size_matching_predicate(
        self, predicate: Predicate, sized: Sized
    ) -> None:
        assert collection.count(predicate)(sized) is True

    @pytest.mark.parametrize(
        "predicate, sized",
        [
            (generic.equal(1), [0, 0]),
            (numeric.le(3), ("a", "b", 1, 2)),
            (numeric.ge(3), "ab"),
        ],
    )
    def test_returns_false_for_size_failing_predicate(
        self, predicate: Predicate, sized: Sized
    ) -> None:
        assert collection.count(predicate)(sized) is False


class TestExists:
    @pytest.mark.parametrize(
        "predicate, iterable",
        [
            (generic.identical(...), ["a", "b", ...]),
            (numeric.greater(0), [-3, 0, 1, 0]),
            (generic.equal("a"), "cbad"),
        ],
    )
    def test_returns_true_for_iterable_containing_satisfying_item(
        self, predicate: Predicate[object], iterable: Iterable
    ) -> None:
        assert collection.exists(predicate)(iterable) is True

    @pytest.mark.parametrize(
        "predicate, iterable",
        [
            (generic.identical(...), ["a", "b", "c"]),
            (numeric.greater(1), [-3, 0, 1, 0]),
            (generic.equal("a"), "cbed"),
        ],
    )
    def test_returns_false_for_iterable_not_containing_satisfying_item(
        self, predicate: Predicate[object], iterable: Iterable
    ) -> None:
        assert collection.exists(predicate)(iterable) is False


class TestEvery:
    @pytest.mark.parametrize(
        "predicate, iterable",
        [
            (generic.identical(1), [1, 1, 1]),
            (generic.identical(1), []),
            (numeric.greater(0), [1, 2]),
            (generic.equal("a"), "aa"),
        ],
    )
    def test_returns_true_for_complete_iterable(
        self, predicate: Predicate[object], iterable: Iterable
    ) -> None:
        assert collection.every(predicate)(iterable) is True

    @pytest.mark.parametrize(
        "predicate, iterable",
        [
            (generic.identical(...), [..., "b", "c"]),
            (numeric.greater(1), [-3, 0, 1, 0, 2]),
            (generic.equal("a"), "abc"),
        ],
    )
    def test_returns_false_for_incomplete_iterable(
        self, predicate: Predicate[object], iterable: Iterable
    ) -> None:
        assert collection.every(predicate)(iterable) is False



from typing import Final

import pytest

from phantom.sized import Empty
from phantom.sized import NonEmpty

parametrize_non_empty: Final = pytest.mark.parametrize(
    "container", ((1,), frozenset({1}), "foo")
)
parametrize_empty: Final = pytest.mark.parametrize("container", ((), frozenset(), ""))
parametrize_mutable: Final = pytest.mark.parametrize("container", ([], set(), {}))


class TestNonEmpty:
    @parametrize_non_empty
    def test_non_empty_container_is_instance(self, container):
        assert isinstance(container, NonEmpty)

    @parametrize_empty
    def test_empty_container_is_instance(self, container):
        assert not isinstance(container, NonEmpty)

    @parametrize_empty
    def test_instantiation_raises_for_empty_container(self, container):
        with pytest.raises(TypeError):
            NonEmpty.parse(container)

    @parametrize_mutable
    def test_instantiation_raises_for_mutable(self, container):
        with pytest.raises(TypeError):
            NonEmpty.parse(container)

    @parametrize_non_empty
    def test_instantiation_returns_instance(self, container):
        assert container is NonEmpty.parse(container)

    def test_subscription_returns_type_alias(self):
        alias = NonEmpty[tuple]
        assert alias.__origin__ is NonEmpty
        (arg,) = alias.__args__
        assert arg is tuple


class TestEmpty:
    @parametrize_non_empty
    def test_non_empty_container_is_instance(self, container):
        assert not isinstance(container, Empty)

    @parametrize_empty
    def test_empty_container_is_instance(self, container):
        assert isinstance(container, Empty)

    @parametrize_non_empty
    def test_instantiation_raises_for_non_empty_container(self, container):
        with pytest.raises(TypeError):
            Empty.parse(container)

    @parametrize_mutable
    def test_instantiation_raises_for_mutable(self, container):
        with pytest.raises(TypeError):
            Empty.parse(container)

    @parametrize_empty
    def test_instantiation_returns_instance(self, container):
        assert container is Empty.parse(container)

    def test_subscription_returns_type_alias(self):
        alias = Empty[frozenset]
        assert alias.__origin__ is Empty
        (arg,) = alias.__args__
        assert arg is frozenset



from typing import Container
from typing import Iterable
from typing import Sized

from phantom.base import Predicate


def contains(value: object) -> Predicate[Container]:
    """Return a predicate that is successful given a container with `value` in it."""

    def compare(container: Container) -> bool:
        return value in container

    return compare


def contained(container: Container) -> Predicate[object]:
    """Return a predicate that is successful given a value contained by `container`."""

    def compare(value: object) -> bool:
        return value in container

    return compare


def count(predicate: Predicate[int]) -> Predicate[Sized]:
    """
    Return a predicate that is successful given an object with a size satisfying
    `predicate`.
    """

    def compare(sized: Sized) -> bool:
        return predicate(len(sized))

    return compare


def exists(predicate: Predicate[object]) -> Predicate[Iterable]:
    """
    Return a predicate that is successful given an iterable where one or more items
    satisfy `predicate`.
    """

    def compare(iterable: Iterable) -> bool:
        return any(predicate(item) for item in iterable)

    return compare


def every(predicate: Predicate[object]) -> Predicate[Iterable]:
    """
    Return a predicate that is successful given an iterable where all items satisfy
    `predicate`.
    """

    def compare(iterable: Iterable) -> bool:
        return all(predicate(item) for item in iterable)

    return compare



import pytest

from phantom import Phantom
from phantom.base import AbstractInstanceCheck
from phantom.base import NotWithinKind
from phantom.base import PhantomMeta
from phantom.predicates.numeric import positive
from phantom.utils import UnresolvedClassAttribute


class TestPhantom:
    def test_subclass_without_predicate_raises(self):
        with pytest.raises(
            UnresolvedClassAttribute, match="must define class attribute __predicate__"
        ):

            class A(Phantom, bound=int):
                ...

    def test_subclass_without_bound_raises(self):
        with pytest.raises(
            UnresolvedClassAttribute, match="must define class attribute __bound__"
        ):

            class A(Phantom, predicate=positive):
                ...

    def test_concrecte_subclass_of_abstract_raises_for_missing_class_attribute(self):
        class A(Phantom, bound=int, abstract=True):
            ...

        with pytest.raises(
            UnresolvedClassAttribute, match="must define class attribute __predicate__"
        ):

            class B(A):
                ...

    def test_can_subclass_without_predicate_if_abstract(self):
        class A(Phantom, bound=int, abstract=True):
            ...

    def test_can_subclass_without_kind(self):
        class A(Phantom, bound=int, predicate=positive):
            ...

        class B(A, bound=str):
            ...

    def test_subclass_outside_kind_raises(self):
        class A(Phantom, kind=int, abstract=True):
            ...

        with pytest.raises(NotWithinKind):

            class B(A, bound=str, abstract=True):
                ...

    def test_can_define_bound_implicitly(self):
        class A(float, Phantom, abstract=True):
            ...

        assert A.__bound__ == (float,)  # type: ignore[misc]

    def test_can_define_bound_explicitly(self):
        class A(Phantom, bound=float, abstract=True):
            ...

        assert A.__bound__ == (float,)  # type: ignore[misc]

    def test_can_inherit_bound(self):
        class A(Phantom, bound=float, abstract=True):
            ...

        class B(A, abstract=True):
            ...

        assert B.__bound__ == (float,)  # type: ignore[misc]

    def test_abstract_instance_check_raises(self):
        class A(Phantom, bound=float, abstract=True):
            ...

        with pytest.raises(AbstractInstanceCheck):
            isinstance(1, A)

    def test_phantom_meta_is_usable_without_phantom_base(self):
        class Alt(metaclass=PhantomMeta):
            ...

        assert isinstance("a", Alt) is False

        class AlwaysTrue(metaclass=PhantomMeta):
            @classmethod
            def __instancecheck__(self, instance: object) -> bool:
                return True

        assert isinstance("a", AlwaysTrue) is True



from typing import Iterable
from typing import Literal

from .base import Predicate
from .base import T


def true(_value: object) -> Literal[True]:
    return True


def false(_value: object) -> Literal[False]:
    return False


def negate(p: Predicate[T]) -> Predicate[T]:
    def check(value: T) -> bool:
        return not p(value)

    return check


def truthy(value: object) -> bool:
    return bool(value)


falsy = negate(truthy)


def both(p: Predicate[T], q: Predicate[T]) -> Predicate[T]:
    def check(value: T) -> bool:
        return p(value) and q(value)

    return check


def all_of(predicates: Iterable[Predicate[T]]) -> Predicate[T]:
    def check(value: T) -> bool:
        return all(p(value) for p in predicates)

    return check


def any_of(predicates: Iterable[Predicate[T]]) -> Predicate[T]:
    def check(value: T) -> bool:
        return any(p(value) for p in predicates)

    return check



from typing import Tuple
from typing import Type
from typing import Union

import pytest

from phantom.predicates import generic


class TestEqual:
    @pytest.mark.parametrize("a, b", [(1, 1.0), (False, 0), (True, 1)])
    def test_returns_true_for_equal_values(self, a: object, b: object) -> None:
        assert generic.equal(a)(b) is True
        assert generic.equal(b)(a) is True

    @pytest.mark.parametrize("a, b", [(2, 1), (1, 1.1), (True, False)])
    def test_returns_false_for_non_equal_values(self, a: object, b: object) -> None:
        assert generic.equal(a)(b) is False
        assert generic.equal(b)(a) is False


class TestIdentical:
    @pytest.mark.parametrize(
        "a, b", [(1, 1), ("a", "a"), (True, True), (False, False), ((), ())]
    )
    def test_returns_true_for_identical_values(self, a: object, b: object) -> None:
        assert generic.identical(a)(b) is True
        assert generic.identical(b)(a) is True

    @pytest.mark.parametrize("a, b", [([], []), (1, 1.0), (False, 0), (True, 1)])
    def test_returns_false_for_different_values(self, a: object, b: object) -> None:
        assert generic.identical(a)(b) is False
        assert generic.identical(b)(a) is False


class TestOfType:
    @pytest.mark.parametrize("instance,types", [(1, int), (1, (int, float))])
    def test_returns_true_for_instance_of_types(
        self, instance: object, types: Union[Type, Tuple[Type, ...]]
    ) -> None:
        assert generic.of_type(types)(instance) is True

    @pytest.mark.parametrize("instance,types", [(1, float), ("", (int, float))])
    def test_returns_false_for_instance_of_other_type(
        self, instance: object, types: Union[Type, Tuple[Type, ...]]
    ) -> None:
        assert generic.of_type(types)(instance) is False



import pytest

from phantom.predicates import numeric


class TestLess:
    def test_returns_true_for_values_below_limit(self) -> None:
        predicate = numeric.less(10)
        assert predicate(9) is True
        assert predicate(9.999) is True
        assert predicate(-5) is True

    def test_returns_false_for_values_above_limit(self) -> None:
        predicate = numeric.less(10)
        assert predicate(10) is False
        assert predicate(11) is False
        assert predicate(123) is False


class TestLE:
    def test_returns_true_for_values_below_limit(self) -> None:
        predicate = numeric.le(10)
        assert predicate(9) is True
        assert predicate(9.999) is True
        assert predicate(-5) is True
        assert predicate(10) is True

    def test_returns_false_for_values_above_limit(self) -> None:
        predicate = numeric.le(10)
        assert predicate(10.0001) is False
        assert predicate(11) is False
        assert predicate(123) is False


class TestGreater:
    def test_returns_true_for_values_above_limit(self) -> None:
        predicate = numeric.greater(120)
        assert predicate(121) is True
        assert predicate(120.0001) is True
        assert predicate(1200) is True

    def test_returns_false_for_values_below_limit(self) -> None:
        predicate = numeric.greater(120)
        assert predicate(120) is False
        assert predicate(119.9999) is False
        assert predicate(-120) is False


class TestGE:
    def test_returns_true_for_values_above_limit(self) -> None:
        predicate = numeric.ge(120)
        assert predicate(121) is True
        assert predicate(120.0001) is True
        assert predicate(1200) is True
        assert predicate(120) is True

    def test_returns_false_for_values_below_limit(self) -> None:
        predicate = numeric.ge(120)
        assert predicate(119.9999) is False
        assert predicate(-120) is False
        assert predicate(119) is False


class TestPositive:
    def test_limits(self) -> None:
        assert numeric.positive(0) is False
        assert numeric.positive(-1) is False
        assert numeric.positive(1) is True
        assert numeric.positive(0.0001) is True


class TestNonPositive:
    def test_limits(self) -> None:
        assert numeric.non_positive(0) is True
        assert numeric.non_positive(-1) is True
        assert numeric.non_positive(1) is False
        assert numeric.non_positive(0.0001) is False


class TestNegative:
    def test_limits(self) -> None:
        assert numeric.negative(0) is False
        assert numeric.negative(-1) is True
        assert numeric.negative(1) is False
        assert numeric.negative(-0.0001) is True


class TestNonNegative:
    def test_limits(self) -> None:
        assert numeric.non_negative(0) is True
        assert numeric.non_negative(-1) is False
        assert numeric.non_negative(1) is True
        assert numeric.non_negative(-0.0001) is False


parametrize_even = pytest.mark.parametrize("value", [-8296, 0, 2, 32, 9314])
parametrize_odd = pytest.mark.parametrize("value", [-13829, -1, 1, 31, 10023])


class TestEven:
    @parametrize_even
    def test_returns_true_for_even_value(self, value: int) -> None:
        assert numeric.even(value) is True

    @parametrize_odd
    def test_returns_false_for_odd_value(self, value: int) -> None:
        assert numeric.even(value) is False


class TestOdd:
    @parametrize_odd
    def test_returns_true_for_odd_value(self, value: int) -> None:
        assert numeric.odd(value) is True

    @parametrize_even
    def test_returns_false_for_even_value(self, value: int) -> None:
        assert numeric.odd(value) is False



import re

from phantom.predicates.re import is_full_match
from phantom.predicates.re import is_match

pattern = re.compile("abc@")
abc_match = is_match(pattern)
abc_full_match = is_full_match(pattern)


class TestIsMatch:
    def test_returns_true_for_matching_string(self) -> None:
        assert abc_match("abc@") is True
        assert abc_match("abc@ extra") is True

    def test_returns_false_for_non_matching_string(self) -> None:
        assert abc_match("abd@") is False
        assert abc_match("extra abc@") is False


class TestIsFullMatch:
    def test_returns_true_for_matching_string(self) -> None:
        assert abc_full_match("abc@") is True

    def test_returns_false_for_non_matching_string(self) -> None:
        assert abc_full_match("abc@ extra") is False
        assert abc_full_match("abd@") is False
        assert abc_full_match("extra abc@") is False



from __future__ import annotations

import abc
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generic
from typing import Iterable
from typing import Optional
from typing import Protocol
from typing import Tuple
from typing import Type
from typing import TypeVar
from typing import Union
from typing import runtime_checkable

from .utils import UnresolvedClassAttribute
from .utils import resolve_class_attr


@runtime_checkable
class InstanceCheckable(Protocol):
    def __instancecheck__(self, instance: object) -> bool:
        ...


class PhantomMeta(abc.ABCMeta):
    """
    Metaclass that defers __instancecheck__ to derived classes and prevents actual
    instance creation.
    """

    def __instancecheck__(self, instance: object) -> bool:
        if not issubclass(self, InstanceCheckable):
            return False
        return self.__instancecheck__(instance)

    # With the current level of metaclass support in mypy it's unlikely that we'll be
    # able to make this context typed, hence the ignores.
    def __call__(cls, instance):  # type: ignore[no-untyped-def]
        return cls.parse(instance)  # type: ignore[attr-defined,misc]


class NotWithinBound(TypeError):
    ...


class NotWithinKind(TypeError):
    ...


def parse_bound(bound: Type[T], instance: object) -> T:
    if not isinstance(instance, bound):
        raise NotWithinBound(
            f"Can't parse {bound.__qualname__} from {type(instance).__name__} value: "
            f"{instance!r}"
        )
    return instance


Derived = TypeVar("Derived")


class PhantomBase(metaclass=PhantomMeta):
    @classmethod
    def parse(cls: Type[Derived], instance: object) -> Derived:
        return parse_bound(cls, instance)

    @classmethod
    @abc.abstractmethod
    def __instancecheck__(cls, instance: object) -> bool:
        ...


class AbstractInstanceCheck(TypeError):
    ...


T = TypeVar("T", covariant=True, bound=object)
TypeSpec = Union[Type[T], Tuple[Type[T], ...]]
Predicate = Callable[[T], bool]


class Phantom(PhantomBase, Generic[T]):
    __predicate__: ClassVar[Predicate[T]]
    __bound__: ClassVar[TypeSpec[T]]
    __kind__: ClassVar[TypeSpec[Any]]
    __abstract__: ClassVar[bool]

    def __init_subclass__(
        cls,
        predicate: Optional[Predicate[T]] = None,
        bound: Optional[Type[T]] = None,
        kind: Optional[TypeSpec] = None,
        abstract: bool = False,
        **kwargs: Any,
    ) -> None:
        super().__init_subclass__(**kwargs)  # type: ignore[call-arg]
        resolve_class_attr(cls, "__abstract__", abstract)
        resolve_class_attr(cls, "__predicate__", predicate)
        resolve_class_attr(cls, "__kind__", kind, required=False)
        cls._resolve_bound(bound)

    @classmethod
    def _interpret_implicit_bound(cls) -> Iterable[type]:
        for type_ in cls.__mro__:
            if type_ is cls:
                continue
            if issubclass(type_, Phantom):
                break
            yield type_
        else:  # pragma: no cover
            raise RuntimeError(f"{cls} is not a subclass of Phantom")

    @classmethod
    def _resolve_bound(cls, class_arg: Optional[TypeSpec[T]]) -> None:
        bound: TypeSpec
        if class_arg is not None:
            bound = class_arg if isinstance(class_arg, tuple) else (class_arg,)
        elif implicit := tuple(cls._interpret_implicit_bound()):
            bound = implicit
        elif (inherited := getattr(cls, "__bound__", None)) is not None:
            bound = inherited
        elif not getattr(cls, "__abstract__", False):
            raise UnresolvedClassAttribute(
                f"Concrete phantom type {cls.__qualname__} must define class attribute "
                f"__bound__."
            )
        else:
            return
        if (kind := getattr(cls, "__kind__", None)) is not None:
            for part in bound if isinstance(bound, Iterable) else (bound,):
                if not issubclass(part, kind):
                    raise NotWithinKind(
                        f"One of the bounds of {cls} ({part}) isn't a subtype of its "
                        f"kind ({kind})"
                    )
        cls.__bound__ = bound

    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        if cls.__abstract__:
            raise AbstractInstanceCheck(
                "Abstract phantom types cannot be used in instance checks"
            )
        return isinstance(instance, cls.__bound__) and cls.__predicate__(instance)



from __future__ import annotations

from typing import Final
from typing import Optional
from typing import cast

import phonenumbers

from phantom import Phantom
from phantom import parse_bound
from phantom.utils import excepts

__all__ = (
    "InvalidPhoneNumber",
    "normalize_phone_number",
    "is_phone_number",
    "is_formatted_phone_number",
    "PhoneNumber",
    "FormattedPhoneNumber",
)


# Ignore due to subclassing Any, since phonenumbers isn't annotated.
class InvalidPhoneNumber(
    phonenumbers.NumberParseException, TypeError  # type: ignore[misc]
):
    INVALID: Final = 99

    def __init__(self, error_type: int = INVALID, msg: str = "Invalid number") -> None:
        super().__init__(error_type, msg)


def _deconstruct_phone_number(
    phone_number: str, country_code: Optional[str] = None
) -> phonenumbers.PhoneNumber:
    try:
        parsed_number = phonenumbers.parse(phone_number, region=country_code)
    except phonenumbers.NumberParseException as e:
        raise InvalidPhoneNumber(e.error_type, e._msg)
    if not phonenumbers.is_valid_number(parsed_number):
        raise InvalidPhoneNumber
    return parsed_number


def normalize_phone_number(
    phone_number: str, country_code: Optional[str] = None
) -> FormattedPhoneNumber:
    normalized = phonenumbers.format_number(
        _deconstruct_phone_number(phone_number, country_code),
        phonenumbers.PhoneNumberFormat.E164,
    )
    return cast(FormattedPhoneNumber, normalized)


is_phone_number = excepts(InvalidPhoneNumber)(_deconstruct_phone_number)


def is_formatted_phone_number(number: str) -> bool:
    try:
        return number == normalize_phone_number(number)
    except InvalidPhoneNumber:
        return False


class PhoneNumber(str, Phantom, predicate=is_phone_number):
    ...


class FormattedPhoneNumber(str, Phantom, predicate=is_formatted_phone_number):
    @classmethod
    def parse(cls, instance: object) -> FormattedPhoneNumber:
        return normalize_phone_number(parse_bound(str, instance))



import pytest

from phantom.interval import Interval
from phantom.interval import Natural
from phantom.interval import NegativeInt
from phantom.interval import Portion


class TestInterval:
    def test_subclassing_without_check_raises(self):
        with pytest.raises(TypeError, match="I must define an interval check$"):

            class I(Interval, abstract=False):  # noqa: E742
                ...


parametrize_negative_ints = pytest.mark.parametrize("i", (-10, -1, -0, +0))
parametrize_positive_ints = pytest.mark.parametrize("i", (10, 1, +0, -0))


class TestNegativeInt:
    @parametrize_negative_ints
    def test_negative_int_is_instance(self, i):
        assert isinstance(i, NegativeInt)

    def test_positive_int_is_not_instance(self):
        assert not isinstance(1, NegativeInt)
        assert not isinstance(10, NegativeInt)

    def test_instantiation_raises_for_positive_int(self):
        with pytest.raises(TypeError):
            NegativeInt.parse(1)
        with pytest.raises(TypeError):
            NegativeInt.parse(10)

    @parametrize_negative_ints
    def test_instantiation_returns_instance(self, i):
        assert i is NegativeInt.parse(i)


class TestNatural:
    @parametrize_positive_ints
    def test_positive_int_is_instance(self, i):
        assert isinstance(i, Natural)

    def test_negative_int_is_not_instance(self):
        assert not isinstance(-1, Natural)
        assert not isinstance(-10, Natural)

    def test_instantiation_raises_for_positive_int(self):
        with pytest.raises(TypeError):
            Natural.parse(-1)
        with pytest.raises(TypeError):
            Natural.parse(-10)
        with pytest.raises(TypeError):
            Natural(-1)

    @parametrize_positive_ints
    def test_instantiation_returns_instance(self, i):
        assert i is Natural.parse(i)
        assert i is Natural(i)


parametrize_portion_values = pytest.mark.parametrize(
    "i", (0.0, 1.0, -0.0, +1.0, 0.8652559794322651)
)
# TODO: Use math.nextafter on Python 3.9 to test as close to limit as possible
parametrize_non_portion_values = pytest.mark.parametrize("i", (-1, -0.1, 1.1, 2))


class TestPortion:
    @parametrize_portion_values
    def test_value_inside_range_is_instance(self, i):
        assert isinstance(i, Portion)

    @parametrize_non_portion_values
    def test_value_outside_range_is_instance(self, i):
        assert not isinstance(i, Portion)

    @parametrize_portion_values
    def test_instantiation_returns_instance(self, i):
        assert i is Portion.parse(i)
        assert i is Portion(i)

    @parametrize_non_portion_values
    def test_instantiation_raises_for_non_portion_values(self, i):
        with pytest.raises(TypeError):
            Portion.parse(i)
        with pytest.raises(TypeError):
            Portion(i)



import pytest

from phantom.boolean import Falsy
from phantom.boolean import Truthy

parametrize_truthy = pytest.mark.parametrize("v", (object(), ("a",), 1, True))
parametrize_falsy = pytest.mark.parametrize("v", ((), 0, False))


class TestTruthy:
    @parametrize_truthy
    def test_truthy_value_is_instance(self, v):
        assert isinstance(v, Truthy)

    @parametrize_falsy
    def test_falsy_value_is_not_instance(self, v):
        assert not isinstance(v, Truthy)

    @parametrize_truthy
    def test_instantiation_returns_instance(self, v):
        assert v is Truthy.parse(v)

    @parametrize_falsy
    def test_instantiation_raises_for_falsy_value(self, v):
        with pytest.raises(TypeError):
            Truthy.parse(v)


class TestFalsy:
    @parametrize_falsy
    def test_falsy_value_is_instance(self, v):
        assert isinstance(v, Falsy)

    @parametrize_truthy
    def test_truthy_value_is_not_instance(self, v):
        assert not isinstance(v, Falsy)

    @parametrize_falsy
    def test_instantiation_returns_instance(self, v):
        assert v is Falsy.parse(v)

    @parametrize_truthy
    def test_instantiation_raises_for_truthy_value(self, v):
        with pytest.raises(TypeError):
            Falsy.parse(v)



from .base import Phantom
from .predicates import boolean


class Truthy(Phantom[object], predicate=boolean.truthy, bound=object):
    ...


class Falsy(Phantom[object], predicate=boolean.falsy, bound=object):
    ...



from __future__ import annotations

from typing import Any
from typing import Optional
from typing import Protocol
from typing import TypeVar

from .base import Phantom
from .base import Predicate
from .predicates import interval
from .utils import resolve_class_attr

N = TypeVar("N", bound=float)


class IntervalCheck(Protocol):
    def __call__(self, a: N, b: N, /) -> Predicate[N]:
        ...


# See issue as to why the numeric tower isn't used for kind here.
# https://github.com/python/mypy/issues/3186
class Interval(Phantom[float], kind=(int, float), abstract=True):
    __check__: IntervalCheck

    def __init_subclass__(
        cls,
        check: Optional[IntervalCheck] = None,
        low: float = float("-inf"),
        high: float = float("inf"),
        **kwargs: Any,
    ) -> None:
        resolve_class_attr(cls, "__check__", check)
        if getattr(cls, "__check__", None) is None:
            raise TypeError(f"{cls.__qualname__} must define an interval check")
        super().__init_subclass__(predicate=cls.__check__(low, high), **kwargs)


class Open(Interval, check=interval.open, abstract=True):
    ...


class Closed(Interval, check=interval.closed, abstract=True):
    ...


class OpenClosed(Interval, check=interval.open_closed, abstract=True):
    ...


class ClosedOpen(Interval, check=interval.closed_open, abstract=True):
    ...


class Natural(int, Open, low=0):
    ...


class NegativeInt(int, Open, high=0):
    ...


class Portion(float, Open, low=0, high=1):
    ...



from typing import Pattern

from phantom.base import Predicate


def is_match(pattern: Pattern[str]) -> Predicate[str]:
    def match(instance: str) -> bool:
        return pattern.match(instance) is not None

    return match


def is_full_match(pattern: Pattern[str]) -> Predicate[str]:
    def full_match(instance: str) -> bool:
        return pattern.fullmatch(instance) is not None

    return full_match



import re

import pytest

from phantom.re import FullMatch
from phantom.re import Match


class A(Match, pattern=re.compile(r"abc")):
    ...


class TestMatch:
    def test_non_matching_string_is_not_instance(self):
        assert not isinstance("abd", A)

    def test_matching_string_is_instance(self):
        assert isinstance("abcd", A)

    def test_instantiation_raises_for_non_matching_string(self):
        with pytest.raises(TypeError):
            A.parse("b")

    def test_instantiation_returns_instance(self):
        s = "abc"
        assert s is A.parse(s)


class B(FullMatch, pattern=re.compile(r"abc")):
    ...


class TestFullMatch:
    def test_non_matching_string_is_not_instance(self):
        assert not isinstance("abcd", B)

    def test_matching_string_is_instance(self):
        assert isinstance("abc", B)

    def test_instantiation_raises_for_non_matching_string(self):
        with pytest.raises(TypeError):
            B.parse("b")

    def test_instantiation_returns_instance(self):
        s = "abc"
        assert s is B.parse(s)



from typing import Any
from typing import Final
from typing import Generic
from typing import Iterable
from typing import MutableMapping
from typing import MutableSequence
from typing import MutableSet
from typing import Protocol
from typing import Sized
from typing import TypeVar
from typing import _ProtocolMeta  # type: ignore[attr-defined]
from typing import runtime_checkable

from .base import Phantom
from .base import PhantomMeta
from .base import Predicate
from .predicates import boolean
from .predicates import collection
from .predicates import generic
from .predicates import numeric

__all__ = (
    "PhantomSized",
    "NonEmpty",
    "Empty",
)


mutable: Final = (MutableSequence, MutableSet, MutableMapping)
T = TypeVar("T", bound=object, covariant=True)


@runtime_checkable
class SizedIterable(Sized, Iterable[T], Protocol[T]):
    ...


# This raises a mypy error because disallow_subclassing_any is enabled and _ProtocolMeta
# isn't publicly typed.
class SizedIterablePhantomMeta(PhantomMeta, _ProtocolMeta):  # type: ignore[misc]
    ...


class PhantomSized(
    SizedIterable[T],
    Phantom[Sized],
    Generic[T],
    metaclass=SizedIterablePhantomMeta,
    bound=SizedIterable,
    abstract=True,
):
    def __init_subclass__(cls, len: Predicate[float], **kwargs: Any) -> None:
        super().__init_subclass__(
            predicate=boolean.both(
                boolean.negate(generic.of_type(mutable)),
                collection.count(len),
            ),
            **kwargs,
        )


class NonEmpty(PhantomSized[T], Generic[T], len=numeric.greater(0)):
    ...


class Empty(PhantomSized[T], Generic[T], len=generic.equal(0)):
    ...
